// xmacros for settings: type, group, key, defautl value
SETTING(string,  Bindings, launch,     DEFAULT_HOTKEY)
SETTING(boolean, Matching, executable, true)
SETTING(boolean, Icons,    show,       true)
SETTING(boolean, Icons,    scale,      true)
SETTING(string,  Border,   color,      "default")
SETTING(integer, Border,   width,      1)
SETTING(string,  Window,   color,      "default")
SETTING(integer, Window,   width,      200)
SETTING(integer, Window,   height,     100)
SETTING(boolean, Window,   round,      true)
SETTING(boolean, Window,   arch,       true)
SETTING(string,  Labels,   color,      "default")
SETTING(integer, Labels,   size1,      14)
SETTING(integer, Labels,   size2,      12)
SETTING(boolean, Labels,   showinput,  true)

